# -*- coding: utf-8 -*-
import os
from ks3.connection import Connection

ak = os.getenv('KS3_IAM_ACCESS_KEY_ID', '<YOUR_ACCESS_KEY>')
sk = os.getenv('KS3_IAM_ACCESS_KEY_SECRET', '<YOUR_SECRET_KEY>')
bucket_name = os.getenv('KS3_BUCKET', '<KS3_TEST_BUCKET>')


def postProcessAndCheck(region=None, key=None):
    endpoint = 'ks3-cn-' + region + '.ksyuncs.com'
    conn = Connection(ak, sk, host=endpoint)
    bucket = conn.get_bucket("a-" + region + "-bucket")
    newKey = bucket.new_key(key)



postProcessAndCheck(region='qingyang-pre', key='static.png')




from ks3.responseResult import ResponseMetadata


class BucketQuota(object):
    def __init__(self, quota=None, *args, **kwargs):
        self.quota = quota

        self.response_metadata = ResponseMetadata(**kwargs)

    def __repr__(self):
        return "quota "+self.quota

    def startElement(self, name, attrs, connection):
        return None

    def endElement(self, name, value, connection):
        if name == 'StorageQuota':
            self.quota = value
        else:
            setattr(self, name, value)


    def to_xml(self):
        s = '<?xml version="1.0" encoding="UTF-8"?>'
        s += '<Quota>'
        if self.quota is not None:
            s += '<StorageQuota>%s</StorageQuota>' % self.quota
        s += '</Quota>'
        return s



"""
This class encapsulates the provider-specific header differences.
"""

import six
from datetime import datetime

import ks3
from ks3.acl import CannedACLStrings as CannedS3ACLStrings
from ks3.acl import Policy
import ks3.exception


HEADER_PREFIX_KEY = 'header_prefix'
METADATA_PREFIX_KEY = 'metadata_prefix'

AWS_HEADER_PREFIX = 'x-amz-'
GOOG_HEADER_PREFIX = 'x-goog-'
KSS_HEADER_PREFIX = 'x-kss-'

ACL_HEADER_KEY = 'acl-header'
AUTH_HEADER_KEY = 'auth-header'
COPY_SOURCE_HEADER_KEY = 'copy-source-header'
COPY_SOURCE_VERSION_ID_HEADER_KEY = 'copy-source-version-id-header'
COPY_SOURCE_RANGE_HEADER_KEY = 'copy-source-range-header'
DELETE_MARKER_HEADER_KEY = 'delete-marker-header'
DATE_HEADER_KEY = 'date-header'
METADATA_DIRECTIVE_HEADER_KEY = 'metadata-directive-header'
RESUMABLE_UPLOAD_HEADER_KEY = 'resumable-upload-header'
SECURITY_TOKEN_HEADER_KEY = 'security-token-header'
STORAGE_CLASS_HEADER_KEY = 'storage-class'
MFA_HEADER_KEY = 'mfa-header'
SERVER_SIDE_ENCRYPTION_KEY = 'server-side-encryption-header'
VERSION_ID_HEADER_KEY = 'version-id-header'
RESTORE_HEADER_KEY = 'restore-header'
TAGGING_COUNT_HEADER_KEY = 'tagging-count-header'
OBJECT_TYPE_HEADER_KEY = 'object-type-header'
APPEND_NEXT_POSITION_HEADER_KEY = 'append-next-position-header'
CHECKSUM_CRC64ECMA_HEADER_KEY = 'checksum-crc64ecma-header'
RETENTION_OVERWRITE_HEADER_KEY = 'retention-overwrite'
RETENTION_ID_HEADER_KEY = 'retention-id'

STORAGE_COPY_ERROR = 'StorageCopyError'
STORAGE_CREATE_ERROR = 'StorageCreateError'
STORAGE_DATA_ERROR = 'StorageDataError'
STORAGE_PERMISSIONS_ERROR = 'StoragePermissionsError'
STORAGE_RESPONSE_ERROR = 'StorageResponseError'
NO_CREDENTIALS_PROVIDED = object()


class ProfileNotFoundError(ValueError):
    pass


class Provider(object):

    CredentialMap = {
        'kss':    ('kss_access_key_id', 'kss_secret_access_key',
                   'kss_security_token', 'kss_profile'),
        'aws':    ('aws_access_key_id', 'aws_secret_access_key',
                   'aws_security_token', 'aws_profile'),
        #'google': ('gs_access_key_id',  'gs_secret_access_key',
        #           None, None),
    }

    AclClassMap = {
        'aws':    Policy,
        #'google': ACL,
        'kss':    Policy
    }

    CannedAclsMap = {
        'aws':    CannedS3ACLStrings,
        #'google': CannedGSACLStrings,
        'kss':    CannedS3ACLStrings
    }

    HostKeyMap = {
        'aws':    's3',
        #'google': 'gs',
        'kss':    's3'
    }

    ChunkedTransferSupport = {
        'aws':    False,
        #'google': True,
        'kss':    False
    }

    MetadataServiceSupport = {
        'aws': True,
        #'google': False,
        'kss': True
    }

    # If you update this map please make sure to put "None" for the
    # right-hand-side for any headers that don't apply to a provider, rather
    # than simply leaving that header out (which would cause KeyErrors).
    HeaderInfoMap = {
        'kss': {
            HEADER_PREFIX_KEY: KSS_HEADER_PREFIX,
            METADATA_PREFIX_KEY: KSS_HEADER_PREFIX + 'meta-',
            ACL_HEADER_KEY: KSS_HEADER_PREFIX + 'acl',
            AUTH_HEADER_KEY: 'AWS',
            COPY_SOURCE_HEADER_KEY: KSS_HEADER_PREFIX + 'copy-source',
            COPY_SOURCE_VERSION_ID_HEADER_KEY: KSS_HEADER_PREFIX +
                                                'copy-source-version-id',
            COPY_SOURCE_RANGE_HEADER_KEY: KSS_HEADER_PREFIX +
                                           'copy-source-range',
            DATE_HEADER_KEY: KSS_HEADER_PREFIX + 'date',
            DELETE_MARKER_HEADER_KEY: KSS_HEADER_PREFIX + 'delete-marker',
            METADATA_DIRECTIVE_HEADER_KEY: KSS_HEADER_PREFIX +
                                            'metadata-directive',
            RESUMABLE_UPLOAD_HEADER_KEY: None,
            SECURITY_TOKEN_HEADER_KEY: KSS_HEADER_PREFIX + 'security-token',
            SERVER_SIDE_ENCRYPTION_KEY: KSS_HEADER_PREFIX +
                                         'server-side-encryption',
            VERSION_ID_HEADER_KEY: KSS_HEADER_PREFIX + 'version-id',
            STORAGE_CLASS_HEADER_KEY: KSS_HEADER_PREFIX + 'storage-class',
            MFA_HEADER_KEY: KSS_HEADER_PREFIX + 'mfa',
            RESTORE_HEADER_KEY: KSS_HEADER_PREFIX + 'restore',
            TAGGING_COUNT_HEADER_KEY: KSS_HEADER_PREFIX + 'tagging-count',
            OBJECT_TYPE_HEADER_KEY: KSS_HEADER_PREFIX + 'object-type',
            APPEND_NEXT_POSITION_HEADER_KEY: KSS_HEADER_PREFIX + 'append-next-position',
            CHECKSUM_CRC64ECMA_HEADER_KEY: KSS_HEADER_PREFIX + 'checksum-crc64ecma',
            RETENTION_OVERWRITE_HEADER_KEY: KSS_HEADER_PREFIX + 'retention-overwrite',
            RETENTION_ID_HEADER_KEY: KSS_HEADER_PREFIX + 'retention-id'
        },
        'aws': {
            HEADER_PREFIX_KEY: AWS_HEADER_PREFIX,
            METADATA_PREFIX_KEY: AWS_HEADER_PREFIX + 'meta-',
            ACL_HEADER_KEY: AWS_HEADER_PREFIX + 'acl',
            AUTH_HEADER_KEY: 'AWS',
            COPY_SOURCE_HEADER_KEY: AWS_HEADER_PREFIX + 'copy-source',
            COPY_SOURCE_VERSION_ID_HEADER_KEY: AWS_HEADER_PREFIX +
                                                'copy-source-version-id',
            COPY_SOURCE_RANGE_HEADER_KEY: AWS_HEADER_PREFIX +
                                           'copy-source-range',
            DATE_HEADER_KEY: AWS_HEADER_PREFIX + 'date',
            DELETE_MARKER_HEADER_KEY: AWS_HEADER_PREFIX + 'delete-marker',
            METADATA_DIRECTIVE_HEADER_KEY: AWS_HEADER_PREFIX +
                                            'metadata-directive',
            RESUMABLE_UPLOAD_HEADER_KEY: None,
            SECURITY_TOKEN_HEADER_KEY: AWS_HEADER_PREFIX + 'security-token',
            SERVER_SIDE_ENCRYPTION_KEY: AWS_HEADER_PREFIX +
                                         'server-side-encryption',
            VERSION_ID_HEADER_KEY: AWS_HEADER_PREFIX + 'version-id',
            STORAGE_CLASS_HEADER_KEY: AWS_HEADER_PREFIX + 'storage-class',
            MFA_HEADER_KEY: AWS_HEADER_PREFIX + 'mfa',
            RESTORE_HEADER_KEY: AWS_HEADER_PREFIX + 'restore',
            TAGGING_COUNT_HEADER_KEY: AWS_HEADER_PREFIX + 'tagging-count'
        },
        'google': {
            HEADER_PREFIX_KEY: GOOG_HEADER_PREFIX,
            METADATA_PREFIX_KEY: GOOG_HEADER_PREFIX + 'meta-',
            ACL_HEADER_KEY: GOOG_HEADER_PREFIX + 'acl',
            AUTH_HEADER_KEY: 'GOOG1',
            COPY_SOURCE_HEADER_KEY: GOOG_HEADER_PREFIX + 'copy-source',
            COPY_SOURCE_VERSION_ID_HEADER_KEY: GOOG_HEADER_PREFIX +
                                                'copy-source-version-id',
            COPY_SOURCE_RANGE_HEADER_KEY: None,
            DATE_HEADER_KEY: GOOG_HEADER_PREFIX + 'date',
            DELETE_MARKER_HEADER_KEY: GOOG_HEADER_PREFIX + 'delete-marker',
            METADATA_DIRECTIVE_HEADER_KEY: GOOG_HEADER_PREFIX  +
                                            'metadata-directive',
            RESUMABLE_UPLOAD_HEADER_KEY: GOOG_HEADER_PREFIX + 'resumable',
            SECURITY_TOKEN_HEADER_KEY: GOOG_HEADER_PREFIX + 'security-token',
            SERVER_SIDE_ENCRYPTION_KEY: None,
            # Note that this version header is not to be confused with
            # the Google Cloud Storage 'x-goog-api-version' header.
            VERSION_ID_HEADER_KEY: GOOG_HEADER_PREFIX + 'version-id',
            STORAGE_CLASS_HEADER_KEY: None,
            MFA_HEADER_KEY: None,
            RESTORE_HEADER_KEY: None,
        }
    }

    ErrorMap = {
        'kss': {
            STORAGE_COPY_ERROR: ks3.exception.S3CopyError,
            STORAGE_CREATE_ERROR: ks3.exception.S3CreateError,
            STORAGE_DATA_ERROR: ks3.exception.S3DataError,
            STORAGE_PERMISSIONS_ERROR: ks3.exception.S3PermissionsError,
            STORAGE_RESPONSE_ERROR: ks3.exception.S3ResponseError,
        },
        'aws': {
            STORAGE_COPY_ERROR: ks3.exception.S3CopyError,
            STORAGE_CREATE_ERROR: ks3.exception.S3CreateError,
            STORAGE_DATA_ERROR: ks3.exception.S3DataError,
            STORAGE_PERMISSIONS_ERROR: ks3.exception.S3PermissionsError,
            STORAGE_RESPONSE_ERROR: ks3.exception.S3ResponseError,
        },
        #'google': {
        #    STORAGE_COPY_ERROR: ks3.exception.GSCopyError,
        #    STORAGE_CREATE_ERROR: ks3.exception.GSCreateError,
        #    STORAGE_DATA_ERROR: ks3.exception.GSDataError,
        #    STORAGE_PERMISSIONS_ERROR: ks3.exception.GSPermissionsError,
        #    STORAGE_RESPONSE_ERROR: ks3.exception.GSResponseError,
        #}
    }

    def __init__(self, name, access_key=None, secret_key=None,
                 security_token=None, profile_name=None):
        self.host = None
        self.port = None
        self.host_header = None
        self.access_key = access_key
        self.secret_key = secret_key
        self.security_token = security_token
        self.profile_name = profile_name
        self.name = name
        self.acl_class = self.AclClassMap[self.name]
        self.canned_acls = self.CannedAclsMap[self.name]
        self._credential_expiry_time = None

        # Load shared credentials file if it exists
        #shared_path = os.path.join(expanduser('~'), '.' + name, 'credentials')
        #self.shared_credentials = Config(do_load=False)
        #if os.path.isfile(shared_path):
        #    self.shared_credentials.load_from_path(shared_path)

        #self.get_credentials(access_key, secret_key, security_token, profile_name)
        self.configure_headers()
        self.configure_errors()

        # Allow config file to override default host and port.
        #host_opt_name = '%s_host' % self.HostKeyMap[self.name]
        #if config.has_option('Credentials', host_opt_name):
        #    self.host = config.get('Credentials', host_opt_name)
        #port_opt_name = '%s_port' % self.HostKeyMap[self.name]
        #if config.has_option('Credentials', port_opt_name):
        #    self.port = config.getint('Credentials', port_opt_name)
        #host_header_opt_name = '%s_host_header' % self.HostKeyMap[self.name]
        #if config.has_option('Credentials', host_header_opt_name):
        #    self.host_header = config.get('Credentials', host_header_opt_name)

#    def get_access_key(self):
#        if self._credentials_need_refresh():
#            self._populate_keys_from_metadata_server()
#        return self._access_key
#
#    def set_access_key(self, value):
#        self._access_key = value
#
#    access_key = property(get_access_key, set_access_key)
#
#    def get_secret_key(self):
#        if self._credentials_need_refresh():
#            self._populate_keys_from_metadata_server()
#        return self._secret_key
#
#    def set_secret_key(self, value):
#        self._secret_key = value
#
#    secret_key = property(get_secret_key, set_secret_key)
#
#    def get_security_token(self):
#        if self._credentials_need_refresh():
#            self._populate_keys_from_metadata_server()
#        return self._security_token
#
#    def set_security_token(self, value):
#        self._security_token = value
#
#    security_token = property(get_security_token, set_security_token)

    def _credentials_need_refresh(self):
        if self._credential_expiry_time is None:
            return False
        else:
            # The credentials should be refreshed if they're going to expire
            # in less than 5 minutes.
            delta = self._credential_expiry_time - datetime.utcnow()
            # python2.6 does not have timedelta.total_seconds() so we have
            # to calculate this ourselves.  This is straight from the
            # datetime docs.
            seconds_left = (
                (delta.microseconds + (delta.seconds + delta.days * 24 * 3600)
                 * 10 ** 6) // 10 ** 6)
            if seconds_left < (5 * 60):
                ks3.log.debug("Credentials need to be refreshed.")
                return True
            else:
                return False

#    def get_credentials(self, access_key=None, secret_key=None,
#                        security_token=None, profile_name=None):
#        access_key_name, secret_key_name, security_token_name, \
#            profile_name_name = self.CredentialMap[self.name]
#
#        # Load profile from shared environment variable if it was not
#        # already passed in and the environment variable exists
#        if profile_name is None and profile_name_name is not None and \
#           profile_name_name.upper() in os.environ:
#            profile_name = os.environ[profile_name_name.upper()]
#
#        shared = self.shared_credentials
#
#        if access_key is not None:
#            self.access_key = access_key
#            ks3.log.debug("Using access key provided by client.")
#        elif access_key_name.upper() in os.environ:
#            self.access_key = os.environ[access_key_name.upper()]
#            ks3.log.debug("Using access key found in environment variable.")
#        elif profile_name is not None:
#            if shared.has_option(profile_name, access_key_name):
#                self.access_key = shared.get(profile_name, access_key_name)
#                ks3.log.debug("Using access key found in shared credential "
#                               "file for profile %s." % profile_name)
#            elif config.has_option("profile %s" % profile_name,
#                                   access_key_name):
#                self.access_key = config.get("profile %s" % profile_name,
#                                             access_key_name)
#                ks3.log.debug("Using access key found in config file: "
#                               "profile %s." % profile_name)
#            else:
#                raise ProfileNotFoundError('Profile "%s" not found!' %
#                                           profile_name)
#        elif shared.has_option('default', access_key_name):
#            self.access_key = shared.get('default', access_key_name)
#            ks3.log.debug("Using access key found in shared credential file.")
#        elif config.has_option('Credentials', access_key_name):
#            self.access_key = config.get('Credentials', access_key_name)
#            ks3.log.debug("Using access key found in config file.")
#
#        if secret_key is not None:
#            self.secret_key = secret_key
#            ks3.log.debug("Using secret key provided by client.")
#        elif secret_key_name.upper() in os.environ:
#            self.secret_key = os.environ[secret_key_name.upper()]
#            ks3.log.debug("Using secret key found in environment variable.")
#        elif profile_name is not None:
#            if shared.has_option(profile_name, secret_key_name):
#                self.secret_key = shared.get(profile_name, secret_key_name)
#                ks3.log.debug("Using secret key found in shared credential "
#                               "file for profile %s." % profile_name)
#            elif config.has_option("profile %s" % profile_name, secret_key_name):
#                self.secret_key = config.get("profile %s" % profile_name,
#                                             secret_key_name)
#                ks3.log.debug("Using secret key found in config file: "
#                               "profile %s." % profile_name)
#            else:
#                raise ProfileNotFoundError('Profile "%s" not found!' %
#                                           profile_name)
#        elif shared.has_option('default', secret_key_name):
#            self.secret_key = shared.get('default', secret_key_name)
#            ks3.log.debug("Using secret key found in shared credential file.")
#        elif config.has_option('Credentials', secret_key_name):
#            self.secret_key = config.get('Credentials', secret_key_name)
#            ks3.log.debug("Using secret key found in config file.")
#        elif config.has_option('Credentials', 'keyring'):
#            keyring_name = config.get('Credentials', 'keyring')
#            try:
#                import keyring
#            except ImportError:
#                ks3.log.error("The keyring module could not be imported. "
#                               "For keyring support, install the keyring "
#                               "module.")
#                raise
#            self.secret_key = keyring.get_password(
#                keyring_name, self.access_key)
#            ks3.log.debug("Using secret key found in keyring.")
#
#        if security_token is not None:
#            self.security_token = security_token
#            ks3.log.debug("Using security token provided by client.")
#        elif ((security_token_name is not None) and
#              (access_key is None) and (secret_key is None)):
#            # Only provide a token from the environment/config if the
#            # caller did not specify a key and secret.  Otherwise an
#            # environment/config token could be paired with a
#            # different set of credentials provided by the caller
#            if security_token_name.upper() in os.environ:
#                self.security_token = os.environ[security_token_name.upper()]
#                ks3.log.debug("Using security token found in environment"
#                               " variable.")
#            elif shared.has_option(profile_name or 'default',
#                                   security_token_name):
#                self.security_token = shared.get(profile_name or 'default',
#                                                 security_token_name)
#                ks3.log.debug("Using security token found in shared "
#                               "credential file.")
#            elif profile_name is not None:
#                if config.has_option("profile %s" % profile_name,
#                                     security_token_name):
#                    ks3.log.debug("config has option")
#                    self.security_token = config.get("profile %s" % profile_name,
#                                                     security_token_name)
#                    ks3.log.debug("Using security token found in config file: "
#                                   "profile %s." % profile_name)
#            elif config.has_option('Credentials', security_token_name):
#                self.security_token = config.get('Credentials',
#                                                 security_token_name)
#                ks3.log.debug("Using security token found in config file.")
#
#        if ((self._access_key is None or self._secret_key is None) and
#                self.MetadataServiceSupport[self.name]):
#            self._populate_keys_from_metadata_server()
#        self._secret_key = self._convert_key_to_str(self._secret_key)

    def _populate_keys_from_metadata_server(self):
        # get_instance_metadata is imported here because of a circular
        # dependency.
        ks3.log.debug("Retrieving credentials from metadata server.")
        from ks3.utils import get_instance_metadata
        timeout = config.getfloat('Boto', 'metadata_service_timeout', 1.0)
        attempts = config.getint('Boto', 'metadata_service_num_attempts', 1)
        # The num_retries arg is actually the total number of attempts made,
        # so the config options is named *_num_attempts to make this more
        # clear to users.
        metadata = get_instance_metadata(
            timeout=timeout, num_retries=attempts,
            data='meta-data/iam/security-credentials/')
        if metadata:
            # I'm assuming there's only one role on the instance profile.
            security = list(metadata.values())[0]
            self._access_key = security['AccessKeyId']
            self._secret_key = self._convert_key_to_str(security['SecretAccessKey'])
            self._security_token = security['Token']
            expires_at = security['Expiration']
            self._credential_expiry_time = datetime.strptime(
                expires_at, "%Y-%m-%dT%H:%M:%SZ")
            ks3.log.debug("Retrieved credentials will expire in %s at: %s",
                           self._credential_expiry_time - datetime.now(), expires_at)

    def _convert_key_to_str(self, key):
        if isinstance(key, six.text_type):
            # the secret key must be bytes and not unicode to work
            #  properly with hmac.new (see http://bugs.python.org/issue5285)
            return str(key)
        return key

    def configure_headers(self):
        header_info_map = self.HeaderInfoMap[self.name]
        self.metadata_prefix = header_info_map[METADATA_PREFIX_KEY]
        self.header_prefix = header_info_map[HEADER_PREFIX_KEY]
        self.acl_header = header_info_map[ACL_HEADER_KEY]
        self.auth_header = header_info_map[AUTH_HEADER_KEY]
        self.copy_source_header = header_info_map[COPY_SOURCE_HEADER_KEY]
        self.copy_source_version_id = header_info_map[
            COPY_SOURCE_VERSION_ID_HEADER_KEY]
        self.copy_source_range_header = header_info_map[
            COPY_SOURCE_RANGE_HEADER_KEY]
        self.date_header = header_info_map[DATE_HEADER_KEY]
        self.delete_marker = header_info_map[DELETE_MARKER_HEADER_KEY]
        self.metadata_directive_header = (
            header_info_map[METADATA_DIRECTIVE_HEADER_KEY])
        self.security_token_header = header_info_map[SECURITY_TOKEN_HEADER_KEY]
        self.resumable_upload_header = (
            header_info_map[RESUMABLE_UPLOAD_HEADER_KEY])
        self.server_side_encryption_header = header_info_map[SERVER_SIDE_ENCRYPTION_KEY]
        self.storage_class_header = header_info_map[STORAGE_CLASS_HEADER_KEY]
        self.version_id = header_info_map[VERSION_ID_HEADER_KEY]
        self.mfa_header = header_info_map[MFA_HEADER_KEY]
        self.restore_header = header_info_map[RESTORE_HEADER_KEY]
        self.tagging_count_header = header_info_map[TAGGING_COUNT_HEADER_KEY]
        self.object_type_header = header_info_map[OBJECT_TYPE_HEADER_KEY]
        self.append_next_position_header = header_info_map[APPEND_NEXT_POSITION_HEADER_KEY]
        self.checksum_crc64ecma_header = header_info_map[CHECKSUM_CRC64ECMA_HEADER_KEY]
        self.retention_overwrite_header = header_info_map[RETENTION_OVERWRITE_HEADER_KEY]
        self.retention_id_header = header_info_map[RETENTION_ID_HEADER_KEY]


    def configure_errors(self):
        error_map = self.ErrorMap[self.name]
        self.storage_copy_error = error_map[STORAGE_COPY_ERROR]
        self.storage_create_error = error_map[STORAGE_CREATE_ERROR]
        self.storage_data_error = error_map[STORAGE_DATA_ERROR]
        self.storage_permissions_error = error_map[STORAGE_PERMISSIONS_ERROR]
        self.storage_response_error = error_map[STORAGE_RESPONSE_ERROR]

    def get_provider_name(self):
        return self.HostKeyMap[self.name]

    def supports_chunked_transfer(self):
        return self.ChunkedTransferSupport[self.name]


# Static utility method for getting default Provider.
def get_default():
    return Provider('kss')



from ks3.objectTagging import Tag
from ks3.responseResult import ResponseMetadata


class BucketLifecycle(object):
    def __init__(self, rule=None, *args, **kwargs):
        if rule is None:
            self.rule = []
        else:
            self.rule = rule

        self.response_metadata = ResponseMetadata(**kwargs)

    def __repr__(self):
        return "rules: " + str(len(self.rule))

    def startElement(self, name, attrs, connection):
        if name == "Rule":
            self.rule.append(Rule())
            return self.rule[-1]
        return None

    def endElement(self, name, value, connection):
        setattr(self, name, value)

    def to_xml(self):
        s = '<LifecycleConfiguration>'
        if self.rule is not None:
            for r in self.rule:
                s += r.to_xml()
        s += '</LifecycleConfiguration>'
        return s


class Rule(object):
    def __init__(self, id=None, filter=None, status=None, expiration=None, transitions=None,
                 noncurrent_version_expiration=None,
                 noncurrent_version_transition=None, abort_incomplete_multipart_upload=None):
        self.id = id
        self.filter = filter
        self.status = status
        self.expiration = expiration
        self.noncurrent_version_expiration = noncurrent_version_expiration
        self.noncurrent_version_transition = noncurrent_version_transition
        self.abort_incomplete_multipart_upload = abort_incomplete_multipart_upload
        if transitions is None:
            self.transitions = []
        else:
            self.transitions = transitions

    def startElement(self, name, attrs, connection):
        if name == 'Expiration':
            self.expiration = Expiration()
            return self.expiration
        if name == 'Filter':
            self.filter = Filter()
            return self.filter
        if name == 'Transition':
            self.transitions.append(Transition())
            return self.transitions[-1]
        if name == 'AbortIncompleteMultipartUpload':
            self.abort_incomplete_multipart_upload = AbortIncompleteMultipartUpload()
        if name == 'NoncurrentVersionTransition':
            self.noncurrent_version_transition = NoncurrentVersionTransition()
            return self.noncurrent_version_transition
        if name == 'NoncurrentVersionExpiration':
            self.noncurrent_version_expiration = NoncurrentVersionExpiration()
            return self.noncurrent_version_expiration
        return None

    def endElement(self, name, value, connection):
        if name == 'ID':
            self.id = value
        elif name == 'Status':
            self.status = value
        else:
            setattr(self, name, value)

    def to_xml(self):
        s = '<Rule>'
        if self.id is not None:
            s += '<ID>%s</ID>' % self.id
        if self.status is not None:
            s += '<Status>%s</Status>' % self.status
        if self.filter is not None:
            s += self.filter.to_xml()
        if self.expiration is not None:
            s += self.expiration.to_xml()
        if self.transitions is not None:
            for t in self.transitions:
                s += t.to_xml()
        if self.abort_incomplete_multipart_upload is not None:
            s += self.abort_incomplete_multipart_upload.to_xml()
        if self.noncurrent_version_expiration is not None:
            s += self.noncurrent_version_expiration.to_xml()
        if self.noncurrent_version_transition is not None:
            s += self.noncurrent_version_transition.to_xml()
        s += '</Rule>'
        return s


class Filter(object):
    def __init__(self, prefix=None, tags=None):
        self.prefix = prefix
        if tags is None:
            self.tags = []
        else:
            self.tags = tags

    def startElement(self, name, attrs, connection):
        if name == "Tag":
            self.tags.append(Tag())
            return self.tags[-1]
        return None

    def endElement(self, name, value, connection):
        if name == 'Prefix':
            self.prefix = value
        else:
            setattr(self, name, value)

    def to_xml(self):
        s = '<Filter>'
        if len(self.tags) != 0:
            s += '<And>'

        if self.prefix is not None:
            s += '<Prefix>%s</Prefix>' % self.prefix

        if len(self.tags) != 0:
            for tag in self.tags:
                s += tag.to_xml()
            s += '</And>'
        s += '</Filter>'
        return s


class Transition(object):
    def __init__(self, days=None, date=None, storage_class=None):
        self.days = days
        self.storage_class = storage_class
        self.date = date

    def startElement(self, name, attrs, connection):
        return None

    def endElement(self, name, value, connection):
        if name == 'Days':
            self.days = value
        elif name == 'StorageClass':
            self.storage_class = value
        elif name == 'Date':
            self.date = value
        else:
            setattr(self, name, value)

    def to_xml(self):
        s = '<Transition>'
        if self.days is not None:
            s += '<Days>%s</Days>' % self.days
        if self.date is not None:
            s += '<Date>%s</Date>' % self.date
        if self.storage_class is not None:
            s += '<StorageClass>%s</StorageClass>' % self.storage_class
        s += '</Transition>'
        return s


class Expiration(object):
    def __init__(self, days=None, date=None):
        self.days = days
        self.date = date

    def startElement(self, name, attrs, connection):
        return None

    def endElement(self, name, value, connection):
        if name == 'Days':
            self.days = value
        elif name == 'Date':
            self.date = value
        else:
            setattr(self, name, value)

    def to_xml(self):
        s = '<Expiration>'
        if self.days is not None:
            s += '<Days>%s</Days>' % self.days
        if self.date is not None:
            s += '<Date>%s</Date>' % self.date
        s += '</Expiration>'
        return s


class NoncurrentVersionTransition(object):
    def __init__(self, noncurrent_days=None, storage_class=None):
        self.noncurrent_days = noncurrent_days
        self.storage_class = storage_class

    def startElement(self, name, attrs, connection):
        return None

    def endElement(self, name, value, connection):
        if name == 'StorageClass':
            self.storage_class = value
        elif name == 'NoncurrentDays':
            self.noncurrent_days = value
        else:
            setattr(self, name, value)

    def to_xml(self):
        s = '<NoncurrentVersionTransition>'
        if self.storage_class is not None:
            s += '<StorageClass>%s</StorageClass>' % self.storage_class
        if self.noncurrent_days is not None:
            s += '<NoncurrentDays>%s</NoncurrentDays>' % self.noncurrent_days
        s += '</NoncurrentVersionTransition>'
        return s


class NoncurrentVersionExpiration(object):
    def __init__(self, noncurrent_days=None):
        self.noncurrent_days = noncurrent_days

    def to_xml(self):
        s = '<NoncurrentVersionExpiration>'
        if self.noncurrent_days is not None:
            s += '<NoncurrentDays>%s</NoncurrentDays>' % self.noncurrent_days
        s += '</NoncurrentVersionExpiration>'
        return s


class AbortIncompleteMultipartUpload(object):
    def __init__(self, days_after_initiation=None, date=None):
        self.days_after_initiation = days_after_initiation
        self.date = date

    def startElement(self, name, attrs, connection):
        return None

    def endElement(self, name, value, connection):
        if name == 'DaysAfterInitiation':
            self.days_after_initiation = value
        elif name == 'Date':
            self.date = value
        else:
            setattr(self, name, value)

    def to_xml(self):
        s = '<AbortIncompleteMultipartUpload>'
        if self.days_after_initiation is not None:
            s += '<DaysAfterInitiation>%s</DaysAfterInitiation>' % self.days_after_initiation
        if self.date is not None:
            s += '<Date>%s</Date>' % self.date
        s += '</AbortIncompleteMultipartUpload>'
        return s



# coding:utf-8
import time
import re
import os
import select

from ks3 import utils
from ks3.utils import get_default_user_agent

try:
    import http.client as httpcli  # for Python 3
    import urllib.parse as parse
    from urllib import parse as urlparse
except ImportError:
    print('python@2.x no longer supported by ks3sdk@2.x, please refer to ks3sdk@1.x for python@2.x support')

from ks3.auth import canonical_string, add_auth_header, url_encode, encode
from ks3.authV4 import add_auth_header as add_auth_header_v4


class CallingFormat:
    PATH = 1
    SUBDOMAIN = 2
    VANITY = 3


class AuthingFormat:
    V2 = 1
    V4 = 2


def merge_meta(headers, metadata):
    final_headers = headers.copy()
    for k in list(metadata.keys()):
        final_headers["x-kss-" + "meta-" + k] = metadata[k]

    return final_headers


def query_args_hash_to_string(query_args):
    pairs = []
    for k, v in list(query_args.items()):
        piece = k
        if v is not None:
            piece += "=%s" % parse.quote_plus(str(v).encode('utf-8'))
            # piece += "=%s" % v
        pairs.append(piece)

    return '&'.join(pairs)


def get_object_url(age, bucket="", key="", secret_access_key="", access_key_id="", query_args={}):
    expire = str(int(time.time()) + age)
    headers = {"Date": expire}
    c_string = canonical_string("GET", bucket, key, query_args, headers)
    path = c_string.split("\n")[-1]

    signature = parse.quote_plus(encode(secret_access_key, c_string))
    if "?" in path:
        url = "http://kss.ksyun.com%s&Expires=%s&AccessKeyId=%s&Signature=%s" % \
              (path, expire, access_key_id, signature)
    else:
        url = "http://kss.ksyun.com%s?Expires=%s&AccessKeyId=%s&Signature=%s" % \
              (path, expire, access_key_id, signature)
    return url


def make_request(server, port, access_key_id, access_key_secret, bucket="", key="", query_args=None, headers=None,
                 data="", metadata=None, method="PUT", call_fmt=CallingFormat.SUBDOMAIN, is_secure=False, domain_mode=False,
                 need_auth_header=True, timeout=10, ua_addon='', block_size=8192):
    if not headers:
        headers = {}
    # if not query_args:
    #    query_args = {}
    if not metadata:
        metadata = {}

    path = ""
    if bucket and not domain_mode:
        if call_fmt == CallingFormat.SUBDOMAIN:
            server = "%s.%s" % (bucket, server)
        elif call_fmt == CallingFormat.VANITY:
            server = bucket
        elif call_fmt == CallingFormat.PATH:
            path += "/%s" % bucket

    path += "/%s" % url_encode(key)
    path = path.replace('//', '/%2F')

    if query_args:
        if isinstance(query_args, dict):
            path += "?" + query_args_hash_to_string(query_args)
        else:
            path += "?" + query_args

    host = "%s:%d" % (server, port)

    if is_secure:
        connection = KS3HTTPSConnection(host=host, blocksize=block_size)
    else:
        connection = KS3HTTPConnection(host=host, blocksize=block_size)

    connection.timeout = timeout
    headers['User-Agent'] = get_default_user_agent() + ' ' + ua_addon
    final_headers = merge_meta(headers, metadata)
    if method == "PUT" and "Content-Length" not in final_headers and not data:
        final_headers["Content-Length"] = "0"
    if method.upper() == "POST" and "Content-Length" not in final_headers and not data:
        final_headers["Content-Length"] = str(len(data))
    if need_auth_header:
        add_auth_header(access_key_id, access_key_secret, final_headers, method,
                        bucket, key, query_args)

    connection.request(method, path, data, final_headers)
    if connection.early_resp:
        resp = connection.early_resp
    else:
        resp = connection.getresponse()
    if 300 <= resp.status < 400:
        loc = resp.getheader('location')
        if loc:
            reg = re.findall('http[s]?://(.*?)(:\d+)?/', loc)
            if reg:
                # 如果返回的是bucket style域名，需要提取region域名出来
                if len(reg[0][0].split('.')) == 4:
                    new_server = reg[0][0].split('.', 1)[1]
                else:
                    new_server = reg[0][0]
                loc_parse = urlparse.urlparse(loc)
                if 'Signature' in loc_parse.query:
                    connection_temp = httpcli.HTTPSConnection(new_server)
                    connection_temp.request('GET', loc_parse.path + '?' + loc_parse.query)
                    try:
                        resp_temp = connection_temp.getresponse()
                        return resp_temp
                    except Exception as err:
                        print(str(err))
                else:
                    if hasattr(data, 'read'):
                        data.seek(0, os.SEEK_SET)
                        if isinstance(data, utils.FpAdapter):
                            data.reset_crc_process()
                    return make_request(new_server, port, access_key_id, access_key_secret, bucket, key, query_args,
                                        headers, data, metadata, method=method, call_fmt=call_fmt, is_secure=is_secure,
                                        need_auth_header=True, ua_addon=ua_addon)
    return resp


# 发送awsv4的请求
def make_request_v4(access_key_id, access_key_secret, method='', service='', region='', query_args=None, headers={},
                    body="", is_secure=False, timeout=10, inner_api=False):
    inner_string = '.inner' if inner_api else ''
    host = service + inner_string + '.api.ksyun.com'

    if (is_secure):
        connection = httpcli.HTTPSConnection(host)
    else:
        connection = httpcli.HTTPConnection(host)
    connection.timeout = timeout

    path = "/"
    if query_args:
        if isinstance(query_args, dict):
            query_args = query_args_hash_to_string(query_args)
    path += "?" + query_args

    headers = add_auth_header_v4(access_key_id, access_key_secret, region, service, host, method, query_args, body,
                                 headers)

    connection.request(method, path, body, headers)
    resp = connection.getresponse()
    return resp


class KS3HTTPConnection(httpcli.HTTPConnection):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.early_resp = None

    def _send_output(self, message_body=None, encode_chunked=False):
        """Send the currently buffered request and clear the buffer.

        Appends an extra \\r\\n to the buffer.
        A message_body may be specified, to be appended to the request.
        """
        self._buffer.extend((b"", b""))
        msg = b"\r\n".join(self._buffer)
        del self._buffer[:]
        self.send(msg)
        if message_body is not None:

            # create a consistent interface to message_body
            if hasattr(message_body, 'read'):
                # Let file-like take precedence over byte-like.  This
                # is needed to allow the current position of mmap'ed
                # files to be taken into account.
                chunks = self._read_readable(message_body)
            else:
                try:
                    # this is solely to check to see if message_body
                    # implements the buffer API.  it /would/ be easier
                    # to capture if PyObject_CheckBuffer was exposed
                    # to Python.
                    memoryview(message_body)
                except TypeError:
                    try:
                        chunks = iter(message_body)
                    except TypeError:
                        raise TypeError("message_body should be a bytes-like "
                                        "object or an iterable, got %r"
                                        % type(message_body))
                else:
                    # the object implements the buffer interface and
                    # can be passed directly into socket methods
                    chunks = (message_body,)

            for chunk in chunks:
                ready_to_read, ready_to_write, _ = select.select([self.sock], [self.sock], [], self.timeout)
                if ready_to_read:
                    self.early_resp = self.getresponse()
                    return
                if not chunk:
                    if self.debuglevel > 0:
                        print('Zero length chunk ignored')
                    continue

                if encode_chunked and self._http_vsn == 11:
                    # chunked encoding
                    chunk = f'{len(chunk):X}\r\n'.encode('ascii') + chunk \
                            + b'\r\n'
                if ready_to_write:
                    try:
                        self.send(chunk)
                    except Exception as e:
                        # if e.errno == errno.EPIPE:
                        #     print("捕获到Broken pipe错误")
                        # 有可能写的时候，发生了BrokenPipeError；为了读取到响应的内容，最后再尝试读取一次
                        if isinstance(e, (BrokenPipeError, ConnectionAbortedError, ConnectionResetError)):
                            last_read_ready, _, _ = select.select([self.sock], [self.sock], [], self.timeout)
                            if last_read_ready:
                                self.early_resp = self.getresponse()
                                return
                            else:
                                raise e
                        else:
                            raise e

            if encode_chunked and self._http_vsn == 11:
                # end chunked transfer
                self.send(b'0\r\n\r\n')


class KS3HTTPSConnection(httpcli.HTTPSConnection, KS3HTTPConnection):
    ...



try:
    import urllib.parse as parse  # for Python 3
    import urllib.request as request
    import urllib.error as error
except ImportError:
    import urllib as parse  # for Python 2
    import urllib as request
    import urllib as error

def fun(bucketname, url="http://10.4.22.2:9911/console"):
    para = parse.urlencode({"bucketname": bucketname})
    path = url + "?%s" % para
    resp = request.urlopen(path)
    return resp.read()

if __name__ == '__main__':
    fun("sdktest123")



# coding=utf-8
import math, os
from ks3.connection import Connection
from filechunkio import FileChunkIO
####################公共头#########################
# 金山云主账号 AccessKey 拥有所有API的访问权限，风险很高。强烈建议您创建并使用子账号账号进行 API 访问或日常运维，请登录 https://uc.console.ksyun.com/pro/iam/#/user/list 创建子账号。
# 通过指定 host(Endpoint)，您可以在指定的地域创建新的存储空间。host(Endpoint) 以北京为例，其它 Region 请按实际情况填写。
conn = Connection('<yourAccessKeyId>', '<yourAccessKeySecret>', host='ks3-cn-beijing.ksyuncs.com')
# 获取存储空间实例
b = conn.get_bucket('<yourBucketName>')
# 源文件路径
source_path = '<yourSourceFilePath>'
# 源文件大小
source_size = os.stat(source_path).st_size

# 初始化分片。获取初始化的 uploadId，之后的操作中将会用到
# 如需在初始化分片时设置文件存储类型，请在 initiate_multipart_upload 中设置相关 headers
# x-kss-storage-class有效值为"STANDARD"、"STANDARD_IA"。"STANDARD"表示标准存储，"STANDARD_IA"表示低频存储，如果不指定，默认为标准存储。
headers = {"x-kss-storage-class": "STANDARD"}
mp = b.initiate_multipart_upload('<yourKeyName>', headers=headers)
# 举例以 50 MiB 为分片大小
chunk_size = 52428800
chunk_count = int(math.ceil(source_size*1.0 / chunk_size*1.0))

# 通过 FileChunkIO 将文件分片
for i in range(chunk_count):
	offset = chunk_size * i
	bytes = min(chunk_size, source_size - offset)
	with FileChunkIO(source_path, 'r', offset=offset, bytes=bytes) as fp:
		# 逐个上传分片
		mp.upload_part_from_file(fp, part_num=i + 1)
# 发送请求，合并分片，完成分片上传
ret = mp.complete_upload()
if ret and ret.status == 200:
	print("上传成功")

# 取消指定upload_id的分片上传事件，已上传的分片会被删除。
# 列出 Bucket 内所有正在进行的分片上传任务
for p in b.get_all_multipart_uploads():
	print(p.id)
	# 取消上传
	print(p.cancel_upload())

# 列出指定上传任务中所有已上传的分片信息
# part_number_marker 指定应该从哪个分片开始列举，只有比设定值大的分片才会被列举
for part in mp.get_all_parts(part_number_marker=2):
	print('part_number:%s' % part.part_number)
	print('request_id:%s' % part.request_id)

# 列举分片上传事件
# 列举 Bucket 中所有的分片上传事件
key_marker='<yourKey>'
for p in b.list_multipart_uploads(key_marker=key_marker):
	print('uploadId:%s,key:%s' % (p.id, p.key_name))
	for i in p:
		print(i.part_number, i.size, i.etag, i.last_modified)



# coding:utf-8
import json

from ks3.exception import KS3ResponseError
from ks3.http import make_request_v4


def get_buckets_data(access_key_id, access_key_secret, start_time=None, end_time=None, action='QueryKs3Data',
                     bucket_names=None, products=None, date_type="Day", inner_api=False, **kwargs):
    """
    :param access_key_id: 金山云提供的ACCESS KEY ID
    :param access_key_secret: 金山云提供的SECRET KEY ID
    """
    query_args = {
        'Action': action,
        'Version': 'v1',
        'StartTime': start_time,
        'EndTime': end_time,
        'DateType': date_type
    }
    # not None and not ""
    if products:
        query_args['Ks3Product'] = products

    if bucket_names:
        query_args['Bucketname'] = bucket_names

    for k, v in list(kwargs.items()):
        if v:
            if k == 'transfer':
                query_args['Transfer'] = v
            elif k == 'request':
                query_args['Request'] = v
            elif k == 'number':
                query_args['Number'] = v
            else:
                query_args[k] = v

    response = make_request_v4(access_key_id, access_key_secret, method='GET', service='ks3bill',
                               region='cn-beijing-6', query_args=query_args, inner_api=inner_api)
    body = response.read()
    if response.status != 200:
        raise KS3ResponseError(response.status, response.reason, body)
    return json.loads(body)


def query_bucket_rank(access_key_id, access_key_secret, start_time=None, end_time=None, date_type="Day", number=10,
                      ks3_product=None, inner_api=False):
    """
    查询bucket使用量排行
    :param access_key_id: 金山云提供的ACCESS KEY ID
    :param access_key_secret: 金山云提供的SECRET KEY ID
    :param end_time: 查询用量开始时间：yyyyMMddHHmm
    :param start_time: 查询用量结束时间（与开始时间同月，不支持跨月查询）：yyyyMMddHHmm
    :param date_type: 支持按天粒度查询，固定值：Day
    :param ks3_product: 可以查询单个或多个统计项，以逗号分隔，如果不填，则查询所有统计项
    :param number: TOP排序的Bucket数量，取值范围为[1-200]
    """
    return get_buckets_data(access_key_id, access_key_secret, start_time=start_time, end_time=end_time, number=number,
                            action='QueryBucketRank', date_type=date_type, products=ks3_product, inner_api=inner_api)


def query_ks3_data(access_key_id, access_key_secret, start_time=None, end_time=None, date_type="Day", bucket_names=None,
                   ks3_product=None, transfer=None, request=None, inner_api=False):
    """
    查询bucket用量详情数据，包括容量、流量、带宽、请求次数、数据取回量、对象标签数等。
    查询业务分析数据，包括以Object、Referer、IP、UA为统计维度的流量或请求次数。

    :param access_key_id: 金山云提供的ACCESS KEY ID
    :param access_key_secret: 金山云提供的SECRET KEY ID
    :param end_time: 查询用量开始时间：yyyyMMddHHmm
    :param start_time: 查询用量结束时间（与开始时间同月，不支持跨月查询）：yyyyMMddHHmm
    :param date_type: 支持按天粒度查询，固定值：Day
    :param bucket_names: 存储空间名称，最多支持同时查询5个存储桶的用量明细，以逗号分隔
    :param ks3_product: 可以查询单个或多个统计项，以逗号分隔，如果不填，则查询所有统计项
    :param transfer: 可以查询单个或多个统计项的流量情况，以逗号分隔。可填参数：Object、Referer、IP、UA，返回TOP200数据
    :param request: 可以查询单个或多个统计项的请求次数情况，以逗号分隔。可填参数：Object、Referer、IP、UA，返回TOP200数据
    """
    return get_buckets_data(access_key_id, access_key_secret, start_time=start_time, end_time=end_time,
                            action='QueryKs3Data', date_type=date_type, bucket_names=bucket_names, products=ks3_product,
                            transfer=transfer, request=request, inner_api=inner_api)





class Prefix(object):
    def __init__(self, bucket=None, name=None):
        self.bucket = bucket
        self.name = name

    def startElement(self, name, attrs, connection):
        return None

    def endElement(self, name, value, connection):
        if name == 'Prefix':
            self.name = value
        else:
            setattr(self, name, value)

    @property
    def provider(self):
        provider = None
        if self.bucket and self.bucket.connection:
            provider = self.bucket.connection.provider
        return provider



import sys

#-----------------------------------------------------------------------------
# Some code below reference to crcmod which base on python2 version
# Replace some functions to compat python3+ version
#
is_py3 = (sys.version_info[0] == 3)
if is_py3:
    xrange = range
    long = int
    sys.maxint = sys.maxsize

#-----------------------------------------------------------------------------
# Export mkCombineFun to user to support crc64 combine feature.
#
# Example:
#
#    import crcmod
#
#    _POLY = 0x142F0E1EBA9EA3693
#    _XOROUT = 0XFFFFFFFFFFFFFFFF
#
#    string_a = '12345'
#    string_b = '67890'
#
#    combine_fun = mkCombineFun(_POLY, 0, True, _XOROUT)
#
#    crc64_a = crcmod.Crc(_POLY, initCrc=0, xorOut=_XOROUT)
#    crc64_a.update(string_a)
#
#    crc64_b = crcmod.Crc(_POLY, initCrc=0, xorOut=_XOROUT)
#    crc64_b.update(string_b)
#
#    combine_fun(crc64_a.crcValue, crc64_b.crcValue, len(string_b))
#

def mkCombineFun(poly, initCrc=~long(0), rev=True, xorOut=0):
    # mask = (1L<<n) - 1

    (sizeBits, initCrc, xorOut) = _verifyParams(poly, initCrc, xorOut)

    mask = (long(1)<<sizeBits) - 1
    if rev:
        poly = _bitrev(long(poly) & mask, sizeBits)
    else:
        poly = long(poly) & mask


    if sizeBits == 64:
        fun = _combine64
    else:
        raise NotImplemented

    def combine_fun(crc1, crc2, len2):
        return fun(poly, initCrc ^ xorOut, rev, xorOut, crc1, crc2, len2)

    return combine_fun


#-----------------------------------------------------------------------------
# The below code implemented crc64 combine logic, the algorithm reference to aliyun-oss-ruby-sdk
# See more details please visist:
#   - https://github.com/aliyun/aliyun-oss-ruby-sdk/tree/master/ext/crcx

GF2_DIM = 64

def gf2_matrix_square(square, mat):
    for n in xrange(GF2_DIM):
        square[n] = gf2_matrix_times(mat, mat[n])


def gf2_matrix_times(mat, vec):
    summary = 0
    mat_index = 0

    while vec:
        if vec & 1:
            summary ^= mat[mat_index]

        vec >>= 1
        mat_index += 1

    return summary


def _combine64(poly, initCrc, rev, xorOut, crc1, crc2, len2):
    if len2 == 0:
        return crc1

    even = [0] * GF2_DIM
    odd = [0] * GF2_DIM

    crc1 ^= initCrc ^ xorOut

    if (rev):
        # put operator for one zero bit in odd
        odd[0] = poly  # CRC-64 polynomial
        row = 1
        for n in xrange(1, GF2_DIM):
            odd[n] = row
            row <<= 1
    else:
        row = 2
        for n in xrange(0, GF2_DIM - 1):
            odd[n] = row
            row <<= 1
        odd[GF2_DIM - 1] = poly

    gf2_matrix_square(even, odd)

    gf2_matrix_square(odd, even)

    while True:
        gf2_matrix_square(even, odd)
        if len2 & long(1):
            crc1 = gf2_matrix_times(even, crc1)
        len2 >>= 1
        if len2 == 0:
            break

        gf2_matrix_square(odd, even)
        if len2 & long(1):
            crc1 = gf2_matrix_times(odd, crc1)
        len2 >>= 1

        if len2 == 0:
            break

    crc1 ^= crc2

    return crc1

#-----------------------------------------------------------------------------
# The below code copy from crcmod, see more detail please visist:
# https://bitbucket.org/cmcqueen1975/crcmod/src/8fb658289c35eff1d37cc47799569f90c5b39e1e/python2/crcmod/crcmod.py?at=default&fileviewer=file-view-default

#-----------------------------------------------------------------------------
# Check the polynomial to make sure that it is acceptable and return the number
# of bits in the CRC.

def _verifyPoly(poly):
    msg = 'The degree of the polynomial must be 8, 16, 24, 32 or 64'
    poly = long(poly) # Use a common representation for all operations
    for n in (8,16,24,32,64):
        low = long(1)<<n
        high = low*2
        if low <= poly < high:
            return n
    raise ValueError(msg)

#-----------------------------------------------------------------------------
# Bit reverse the input value.

def _bitrev(x, n):
    x = long(x)
    y = long(0)
    for i in xrange(n):
        y = (y << 1) | (x & long(1))
        x = x >> 1
    if ((long(1)<<n)-1) <= sys.maxint:
        return int(y)
    return y

#-----------------------------------------------------------------------------
# The following function validates the parameters of the CRC, namely,
# poly, and initial/final XOR values.
# It returns the size of the CRC (in bits), and "sanitized" initial/final XOR values.

def _verifyParams(poly, initCrc, xorOut):
    sizeBits = _verifyPoly(poly)

    mask = (long(1)<<sizeBits) - 1

    # Adjust the initial CRC to the correct data type (unsigned value).
    initCrc = long(initCrc) & mask
    if mask <= sys.maxint:
        initCrc = int(initCrc)

    # Similar for XOR-out value.
    xorOut = long(xorOut) & mask
    if mask <= sys.maxint:
        xorOut = int(xorOut)

    return (sizeBits, initCrc, xorOut)



"""
Some handy utility functions used by several classes.
"""
import crcmod

import ks3
import six
import io
import base64

from hashlib import md5

from ks3.compat import encodebytes
from ks3.crc64_combine import mkCombineFun
from ks3.exception import KS3ClientError

try:
    import urllib.parse as parse  # for Python 3
except ImportError:
    import urllib as parse  # for Python 2


def get_utf8_value(value):
    if not six.PY2 and isinstance(value, bytes):
        return value

    if not isinstance(value, six.string_types):
        value = six.text_type(value)

    if isinstance(value, six.text_type):
        value = value.encode('utf-8')

    return value


def merge_headers_by_name(name, headers):
    """
    Takes a specific header name and a dict of headers {"name": "value"}.
    Returns a string of all header values, comma-separated, that match thepyo
    input header name, case-insensitive.

    """
    matching_headers = find_matching_headers(name, headers)
    return ','.join(str(headers[h]) for h in matching_headers
                    if headers[h] is not None)


def find_matching_headers(name, headers):
    """
    Takes a specific header name and a dict of headers {"name": "value"}.
    Returns a list of matching header names, case-insensitive.

    """
    return [h for h in headers if h.lower() == name.lower()]


def merge_meta(headers, metadata, provider=None):
    if not provider:
        provider = ks3.provider.get_default()
    metadata_prefix = provider.metadata_prefix
    final_headers = headers.copy()
    for k in list(metadata.keys()):
        if k.lower() in ks3.key.Key.base_user_settable_fields:
            final_headers[k] = metadata[k]
        else:
            final_headers[metadata_prefix + k] = metadata[k]

    return final_headers


def compute_md5(fp, buf_size=8192, size=None):
    """
    Compute MD5 hash on passed file and return results in a tuple of values.
    """
    return compute_hash(fp, buf_size, size, hash_algorithm=md5)


def compute_hash(fp, buf_size=8192, size=None, hash_algorithm=md5):
    hash_obj = hash_algorithm()
    spos = fp.tell()
    if size and size < buf_size:
        s = fp.read(size)
    else:
        s = fp.read(buf_size)
    while s:
        if not isinstance(s, bytes):
            s = s.encode('utf-8')
        hash_obj.update(s)
        if size:
            size -= len(s)
            if size <= 0:
                break
        if size and size < buf_size:

            s = fp.read(size)
        else:
            s = fp.read(buf_size)
    hex_digest = hash_obj.hexdigest()
    base64_digest = encodebytes(hash_obj.digest()).decode('utf-8')
    if base64_digest[-1] == '\n':
        base64_digest = base64_digest[0:-1]
    # data_size based on bytes read.
    data_size = fp.tell() - spos
    fp.seek(spos)
    return (hex_digest, base64_digest, data_size)


def compute_encrypted_md5(fp, buf_size=8192, hash_algorithm=md5):
    hash_obj = hash_algorithm()
    s = fp.read(buf_size)
    while s:
        if not isinstance(s, bytes):
            s = s.encode('utf-8')
        hash_obj.update(s)
        s = fp.read(buf_size)
    hex_digest = hash_obj.hexdigest()
    base64_digest = encodebytes(hash_obj.digest()).decode('utf-8')
    if base64_digest[-1] == '\n':
        base64_digest = base64_digest[0:-1]
    # data_size based on bytes read.
    SEEK_SET = getattr(io, 'SEEK_SET', 0)
    fp.seek(SEEK_SET)
    return (hex_digest, base64_digest)


def convert_adp_headers(adps):
    if adps:
        fop = ""
        for op in adps:
            fop = "%s|tag=saveas" % op["command"]
            if op["bucket"]:
                fop += "&bucket=%s" % (op["bucket"])
            if op["key"]:
                fop += "&object=%s" % (base64.b64encode(op["key"]))
            fop = "%s;" % fop
        fop = fop.rstrip(";")
        headers = {"kss-async-process": parse.quote(fop),
                   "kss-notifyurl": parse.quote("http://127.0.0.1:9000/notify/url")}
        return headers
    else:
        return None


def compute_base64_md5_digest(data):
    m = md5()
    m.update(data)
    base64_digest = encodebytes(m.digest()).decode('utf-8')
    if base64_digest[-1] == '\n':
        base64_digest = base64_digest[0:-1]
    return base64_digest


def get_default_user_agent():
    # import platform
    # platform.version()
    # platform.version()
    return 'PythonSDK/' + ks3.__version__


def to_boolean(value, true_value='true'):
    if value == true_value:
        return True
    else:
        return False


class Crc64(object):
    _POLY = 0x142F0E1EBA9EA3693
    _XOROUT = 0XFFFFFFFFFFFFFFFF

    def __init__(self, init_crc=0):
        self.crc64 = crcmod.Crc(self._POLY, initCrc=init_crc, rev=True, xorOut=self._XOROUT)

        self.crc64_combineFun = mkCombineFun(self._POLY, initCrc=init_crc, rev=True, xorOut=self._XOROUT)

    def __call__(self, data):
        self.update(data)

    def update(self, data):
        self.crc64.update(data)

    def combine(self, crc1, crc2, len2):
        return self.crc64_combineFun(crc1, crc2, len2)

    @property
    def crc(self):
        return self.crc64.crcValue


def compute_file_crc64(file_name, start=None, end=None, block_size=64 * 1024):
    with open(file_name, 'rb') as fp:
        fp = FpAdapter(fp)
        if start is not None:
            fp.seek(start)
        while True:
            if end is not None:
                if fp.tell() >= end:
                    break
                else:
                    data = fp.read(min(block_size, end - fp.tell() + 1))
            else:
                data = fp.read(block_size)
            if not data:
                break

    return str(fp.crc)


def compute_data_crc64(data, init_crc=0):
    """
    Calculate the crc64 of a string

    :param data: The content of the string
    :param init_crc: The initial value of crc64, default is 0
    :return The crc64 value of the string
    """
    if not isinstance(data, bytes):
        data = data.encode('utf-8')
    crc64 = Crc64(init_crc)
    crc64.update(data)
    return str(crc64.crc)


class FpAdapter(object):
    def __init__(self, fp):
        super(FpAdapter, self).__init__()
        self.fp = fp
        self.crc64_handler = Crc64()
        self.first_read_done = False
        self.counter = 0

    def read(self, size=0):
        try:
            data = self.fp.read(size)
            self.counter += len(data) if data else 0
        except Exception as e:
            raise KS3ClientError('Read file error: fileName:%s, readSize:%d, readOffset: %d, error:%s' % (self.fp.name, size, self.counter, e))
        if data and not self.first_read_done:
            self.crc64_handler.update(data)
        else:
            self.first_read_done = True
        return data

    @property
    def crc(self):
        return self.crc64_handler.crc

    @property
    def name(self):
        return self.fp.name

    def reset_crc_process(self):
        self.crc64_handler = Crc64()
        self.first_read_done = False
        self.counter = 0

    def seek(self, *args, **kwargs):
        return self.fp.seek(*args, **kwargs)

    def tell(self, *args, **kwargs):
        return self.fp.tell(*args, **kwargs)

    def close(self):
        self.fp.close()

    def __len__(self):
        return len(self.fp)

    def __str__(self):
        return str(self.fp)

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


# return if client_crc == server_crc
def check_crc(client_crc, server_crc):
    if client_crc is None or server_crc is None:
        return True
    return client_crc is not None and server_crc is not None and client_crc == server_crc



# -*- coding: utf-8 -*-
from ks3.responseResult import ResponseMetadata


class BucketCors(object):
    def __init__(self, rules=None, *args, **kwargs):
        if rules is None:
            self.rules = []
        else:
            self.rules = rules

        self.response_metadata = ResponseMetadata(**kwargs)

    def __repr__(self):
        return "rules: " + str(len(self.rules))

    def startElement(self, name, attrs, connection):
        if name == "CORSRule":
            self.rules.append(CORSRule())
            return self.rules[-1]
        return None

    def endElement(self, name, value, connection):
        setattr(self, name, value)

    def to_xml(self):
        # caller is responsible to encode to utf-8
        s = u'<CORSConfiguration xmlns="http://s3.amazonaws.com/doc/2006-03-01/">'
        if self.rules:
            for rule in self.rules:
                s += rule.to_xml()
        s += '</CORSConfiguration>'
        return s


class CORSRule(object):
    def __init__(self, origins=None, methods=None, headers=None, max_age=None, exposed_headers=None):
        if origins is None:
            self.origins = []
        else:
            self.origins = origins
        if methods is None:
            self.methods = []
        else:
            self.methods = methods
        if headers is None:
            self.headers = []
        else:
            self.headers = headers
        if exposed_headers is None:
            self.exposed_headers = []
        else:
            self.exposed_headers = exposed_headers
        self.max_age = max_age

    def startElement(self, name, attrs, connection):
        return None

    def endElement(self, name, value, connection):
        if name == "AllowedOrigin":
            self.origins.append(value)
        if name == "AllowedMethod":
            self.methods.append(value)
        if name == "AllowedHeader":
            self.headers.append(value)
        if name == "ExposeHeader":
            self.exposed_headers.append(value)
        if name == 'MaxAgeSeconds':
            self.max_age = value
        else:
            setattr(self, name, value)

    def to_xml(self):
        s = '<CORSRule>'
        if self.origins is not None:
            for origin in self.origins:
                s += '<AllowedOrigin>%s</AllowedOrigin>' % origin
        if self.methods is not None:
            for method in self.methods:
                s += '<AllowedMethod>%s</AllowedMethod>' % method
        if self.headers is not None:
            for header in self.headers:
                s += '<AllowedHeader>%s</AllowedHeader>' % header
        if self.max_age is not None:
            s += '<MaxAgeSeconds>%s</MaxAgeSeconds>' % self.max_age
        if self.exposed_headers is not None:
            for exposed_header in self.exposed_headers:
                s += '<ExposeHeader>%s</ExposeHeader>' % exposed_header
        s += '</CORSRule>'
        return s


try:
    import urllib.parse as parse  # for Python 3
except ImportError:
    import urllib as parse  # for Python 2


def bucket_lister(bucket, marker='', continuation_token='', **params):
    """
    A generator function for listing keys in a bucket.
    """
    more_results = True
    k = None
    while more_results:
        rs = bucket.get_all_keys(marker=marker, continuation_token=continuation_token, **params)
        for k in rs:
            yield k

        if rs.next_continuation_token:
            continuation_token = rs.next_continuation_token
        elif k:
            marker = rs.next_marker or k.name
        if marker:
            '''if isinstance(marker, six.text_type):
                marker = marker.encode('utf-8')'''
            marker = parse.unquote(marker)
        more_results = rs.is_truncated


class BucketListResultSet(object):
    """
    A resultset for listing keys within a bucket.  Uses the bucket_lister
    generator function and implements the iterator interface.  This
    transparently handles the results paging from S3 so even if you have
    many thousands of keys within the bucket you can iterate over all
    keys in a reasonably efficient manner.
    """

    def __init__(self, bucket=None, prefix='', delimiter='', marker='', max_keys='', encoding_type='',
                 continuation_token='', fetch_owner='', list_type=None, start_after=''):
        self.bucket = bucket
        self.prefix = prefix
        self.delimiter = delimiter
        self.marker = marker
        self.max_keys = max_keys
        self.encoding_type = encoding_type
        self.continuation_token = continuation_token
        self.fetch_owner = fetch_owner
        self.list_type = list_type
        self.start_after = start_after

    def __iter__(self):
        return bucket_lister(self.bucket, prefix=self.prefix, delimiter=self.delimiter, marker=self.marker,
                             max_keys=self.max_keys, encoding_type=self.encoding_type,
                             continuation_token=self.continuation_token, fetch_owner=self.fetch_owner,
                             list_type=self.list_type, start_after=self.start_after)


def versioned_bucket_lister(bucket, prefix='', delimiter='',
                            key_marker='', version_id_marker='', headers=None,
                            encoding_type=None):
    """
    A generator function for listing versions in a bucket.
    """
    more_results = True
    k = None
    while more_results:
        rs = bucket.get_all_versions(prefix=prefix, key_marker=key_marker,
                                     version_id_marker=version_id_marker,
                                     delimiter=delimiter, headers=headers,
                                     max_keys=999, encoding_type=encoding_type)
        for k in rs:
            yield k
        key_marker = rs.next_key_marker
        if key_marker:
            key_marker = parse.unquote(key_marker)
        version_id_marker = rs.next_version_id_marker
        more_results = rs.is_truncated


class VersionedBucketListResultSet(object):
    """
    A resultset for listing versions within a bucket.  Uses the bucket_lister
    generator function and implements the iterator interface.  This
    transparently handles the results paging from S3 so even if you have
    many thousands of keys within the bucket you can iterate over all
    keys in a reasonably efficient manner.
    """

    def __init__(self, bucket=None, prefix='', delimiter='', key_marker='',
                 version_id_marker='', headers=None, encoding_type=None):
        self.bucket = bucket
        self.prefix = prefix
        self.delimiter = delimiter
        self.key_marker = key_marker
        self.version_id_marker = version_id_marker
        self.headers = headers
        self.encoding_type = encoding_type

    def __iter__(self):
        return versioned_bucket_lister(self.bucket, prefix=self.prefix,
                                       delimiter=self.delimiter,
                                       key_marker=self.key_marker,
                                       version_id_marker=self.version_id_marker,
                                       headers=self.headers,
                                       encoding_type=self.encoding_type)


def bucket_retention_lister(bucket, prefix='', delimiter='', marker='', max_keys=''):
    more_results = True
    k = None
    while more_results:
        rs = bucket.get_all_retention_keys(prefix=prefix, delimiter=delimiter, marker=marker, max_keys=max_keys)
        for k in rs:
            yield k
        if k:
            marker = rs.next_marker or k.name
        if marker:
            marker = parse.unquote(marker)
        more_results = rs.is_truncated


class BucketRetentionListResultSet(object):
    def __init__(self, bucket=None, prefix='', delimiter='', marker='', max_keys=''):
        self.bucket = bucket
        self.prefix = prefix
        self.delimiter = delimiter
        self.marker = marker
        self.max_keys = max_keys

    def __iter__(self):
        return bucket_retention_lister(self.bucket, prefix=self.prefix,
                                       delimiter=self.delimiter, marker=self.marker, max_keys=self.max_keys)


def multipart_upload_lister(bucket, key_marker='',
                            upload_id_marker='',
                            headers=None, encoding_type=None):
    """
    A generator function for listing multipart uploads in a bucket.
    """
    more_results = True
    k = None
    while more_results:
        rs = bucket.get_all_multipart_uploads(key_marker=key_marker,
                                              upload_id_marker=upload_id_marker,
                                              headers=headers,
                                              encoding_type=encoding_type)
        for k in rs:
            yield k
        key_marker = rs.next_key_marker
        upload_id_marker = rs.next_upload_id_marker
        more_results = rs.is_truncated


class MultiPartUploadListResultSet(object):
    """
    A resultset for listing multipart uploads within a bucket.
    Uses the multipart_upload_lister generator function and
    implements the iterator interface.  This
    transparently handles the results paging from S3 so even if you have
    many thousands of uploads within the bucket you can iterate over all
    keys in a reasonably efficient manner.
    """

    def __init__(self, bucket=None, key_marker='',
                 upload_id_marker='', headers=None, encoding_type=None):
        self.bucket = bucket
        self.key_marker = key_marker
        self.upload_id_marker = upload_id_marker
        self.headers = headers
        self.encoding_type = encoding_type

    def __iter__(self):
        return multipart_upload_lister(self.bucket,
                                       key_marker=self.key_marker,
                                       upload_id_marker=self.upload_id_marker,
                                       headers=self.headers,
                                       encoding_type=self.encoding_type)



# coding:utf-8

def set_response_func(self, **kwargs):
    self.response_metadata = dict()
    if "headers" in kwargs:
        self.response_metadata["headers"] = kwargs["headers"]
        self.response_metadata["request_id"] = kwargs["headers"].get("x-kss-request-id")
    if "status" in kwargs:
        self.response_metadata["status"] = kwargs["status"]
    if "reason" in kwargs:
        self.response_metadata["reason"] = kwargs["reason"]
    if "raw_body" in kwargs:
        self.response_metadata["raw_body"] = kwargs["raw_body"]


class ResponseResult(object):
    def __init__(self, data, **kwargs):
        self.data = data
        self.response_metadata = ResponseMetadata(**kwargs)

    @property
    def status(self):
        return self.response_metadata.status

    @property
    def headers(self):
        return self.response_metadata.headers

    def getheader(self, item, default=None):
        # 避免使用data里的属性
        # return self.data.getheader(item)
        headers = self.response_metadata.headers.get_all(item) or default
        if isinstance(headers, str) or not hasattr(headers, '__iter__'):
            return headers
        else:
            return ', '.join(headers)

    @property
    def msg(self):
        return self.headers

    @property
    def request_id(self):
        return self.response_metadata.request_id

    def __repr__(self):
        return 'ResponseResult, status: %s, reason: %s, request_id: %s' % (self.response_metadata.status,
                                                                           self.response_metadata.reason,
                                                                           self.response_metadata.request_id)


class IterableResponseResult(ResponseResult):
    def __iter__(self):
        return iter(self.data)

    def __next__(self):
        return next(self.data)

    def __len__(self):
        return len(self.data)

    def __getitem__(self, index):
        return self.data[index]

    def __setitem__(self, index, value):
        self.data[index] = value

    def __delitem__(self, index):
        del self.data[index]

    def insert(self, index, value):
        self.data.insert(index, value)


class ResponseMetadata(object):
    def __init__(self, **kwargs):
        if "headers" in kwargs:
            self.headers = kwargs["headers"]
            self.request_id = kwargs["headers"].get("x-kss-request-id")
        if "status" in kwargs:
            self.status = kwargs["status"]
        if "reason" in kwargs:
            self.reason = kwargs["reason"]
        if "raw_body" in kwargs:
            self.raw_body = kwargs["raw_body"]
